/* PICOBOOT protocol is in my rp2040 datasheet starting at pg148 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <libusb.h>

#define MIN(_x,_y) (((_x) < (_y)) ? (_x) : (_y))

static int elf_endian;
static unsigned elf_entry;		/* 0 for FLASH or entry PC for SRAM */
static unsigned elf_phoff, elf_phnum;

#define SRAM_START 0x20000000
#define SRAM_LEN 0x42000
#define XIP_SRAM_START 0x15000000
#define XIP_SRAM_LEN 0x4000
#define FLASH_START 0x10000000
#define FLASH_LEN 0x200000	/* assume 2MB of flash, even though the address
				 * range extends all the way to 0x15000000 */
/* PICOBOOT supposedly supports writing to XIP SRAM, but this tool doesn't */
#define IS_SRAM_ADDR(_addr) \
	(((_addr) >= SRAM_START) && ((_addr) < (SRAM_START+SRAM_LEN)))
#define IS_FLASH_ADDR(_addr) \
	(((_addr) >= FLASH_START) && ((_addr) < (FLASH_START+FLASH_LEN)))

#define BLKSZ 256			/* flash writes must be 256 aligned */
#define FLASH_SECTOR_SIZE 4096		/* flash erases must be 4k aligned */


struct segment_rec {
	struct segment_rec *next;
	unsigned fileofs, addr, len;
};
static struct segment_rec *segments = NULL, *segments_last = NULL;

static void
new_segment(unsigned fileofs, unsigned addr, unsigned len)
{
	struct segment_rec *n = malloc(sizeof *n);
	n->fileofs = fileofs;
	n->addr = addr;
	n->len = len;
	n->next = NULL;
	if (segments_last) {
		segments_last->next = n;
	} else {
		segments = n;
	}
	segments_last = n;
}


static unsigned
elf_4byte(unsigned char *buf)
{
	return ((elf_endian == 1)
		? ((((unsigned)buf[0]) << 0) |		/* little-endian */
		   (((unsigned)buf[1]) << 8) |
		   (((unsigned)buf[2]) << 16) |
		   (((unsigned)buf[3]) << 24))
		: ((((unsigned)buf[0]) << 24) |		/* big-endian */
		   (((unsigned)buf[1]) << 16) |
		   (((unsigned)buf[2]) << 8) |
		   (((unsigned)buf[3]) << 0)));
}

static unsigned
elf_2byte(unsigned char *buf)
{
	return ((elf_endian == 1)
		? ((((unsigned)buf[0]) << 0) |		/* little-endian */
		   (((unsigned)buf[1]) << 8))
		: ((((unsigned)buf[0]) << 8) |		/* big-endian */
		   (((unsigned)buf[1]) << 0)));
}

static int
elf_read_header(FILE *f)
{
	unsigned char buf[0x34];
	if (fread(buf, sizeof buf, 1, f) != 1) { perror("fread"); return -1; }
	if (memcmp(buf, "\x7f" "ELF", 4)) { fprintf(stderr, "not ELF\n"); return -1; }
	if (buf[4] != 1) { fprintf(stderr, "not 32-bit\n"); return -1; }
	if (buf[6] != 1) { fprintf(stderr, "not ELF 1.0\n"); return -1; }
	elf_endian = buf[5];
	if (elf_2byte(&buf[0x12]) != 0x28) { fprintf(stderr, "not ARM\n"); return -1; }
	elf_entry = elf_4byte(&buf[0x18]);
	int phentsize = elf_2byte(&buf[0x2a]);
	if (phentsize != 0x20) { fprintf(stderr, "invalid phentsize\n"); return -1; }
	elf_phoff = elf_4byte(&buf[0x1c]);
	elf_phnum = elf_2byte(&buf[0x2c]);
	return 0;
}

/* read a single ELF program header */
static int
elf_read_ph(FILE *f)
{
	unsigned char buf[0x20];
	if (fread(buf, sizeof buf, 1, f) != 1) { perror("fread"); return -1; }
	if (elf_4byte(&buf[0]) != 1) return 0;	/* ignore non-PT_LOAD */
	int memsz = elf_4byte(&buf[0x14]);
	int filesz = elf_4byte(&buf[0x10]);
	if (memsz > filesz) memsz = filesz;	/* ignore uninit portion */
	if (!memsz) return 0;			/* ignore empty */
	unsigned offset = elf_4byte(&buf[0x04]);/* file offset */
	unsigned paddr = elf_4byte(&buf[0x0c]);	/* target phys addr */
	new_segment(offset, paddr, memsz);
	return 0;
}

static int
elf_read_phs(FILE *f)
{
	int i;
	if (fseek(f, elf_phoff, SEEK_SET)) { perror("fseek phoff"); return -1; }
	for (i = 0; i < elf_phnum; i++) {
		if (elf_read_ph(f)) { return -1; }
	}
	return 0;
}


/* split output into BLKSZ-byte blocks (PICOBOOT requirement for FLASH) */
static unsigned char **blocks;
static unsigned blocks_start;		/* addr of first block */
static int blocks_count;
static int
accum_block(unsigned addr, unsigned char *buf, unsigned len)
{
	if ((addr < blocks_start) ||
	    ((addr+len) > blocks_start+(blocks_count*BLKSZ))) {
		fprintf(stderr, "PT_LOAD segment at %08X not in %s",
			addr, (elf_entry ? "SRAM" : "FLASH"));
		return -1;
	}

	addr -= blocks_start;
	while (len) {
		unsigned bn = addr / BLKSZ;
		unsigned bo = addr - (bn*BLKSZ);
		unsigned bl = MIN(BLKSZ-bo, len);
		if (!blocks[bn]) {
			blocks[bn] = malloc(BLKSZ);
			memset(blocks[bn], 0, BLKSZ);
		}
		memcpy(blocks[bn]+bo, buf, bl);
		buf += bl;
		addr += bl;
		len -= bl;
	}
	return 0;
}

static int
accum_blocks(FILE *f)
{
	struct segment_rec *s;

	if (elf_entry) {		/* must be SRAM */
		blocks_start = SRAM_START;
		blocks_count = SRAM_LEN/BLKSZ;
	} else {
		blocks_start = FLASH_START;
		blocks_count = FLASH_LEN/BLKSZ;
	}
	blocks = malloc(blocks_count * sizeof *blocks);
	memset(blocks, 0, blocks_count * sizeof *blocks);

	for (s = segments; s; s = s->next) {
		if (fseek(f, s->fileofs, SEEK_SET)) {
			perror("segment fseek");
			return -1;
		}
		unsigned char *buf = malloc(s->len);
		if (fread(buf, s->len, 1, f) != 1) {
			perror("segment fread");
			return -1;
		}
		if (accum_block(s->addr, buf, s->len)) return -1;
		free(buf);
	}
	return 0;
}


/* The RP2040 PICOBOOT interface is all little-endian (which is nice, the
 * mass-storage interface is both endians). */
#define tobuf_4byte_le(dst, src) { \
	unsigned __src = (src); \
	unsigned char *__dst = (unsigned char *)(dst); \
	__dst[0] = ((__src)&0xFF); \
	__dst[1] = ((__src>>8)&0xFF); \
	__dst[2] = ((__src>>16)&0xFF); \
	__dst[3] = ((__src>>24)&0xFF); }

#define RP2040_VENDOR 0x2e8a		/* Raspberry Pi foundation */
#define RP2040_PRODUCT 0x0003		/* RP2040 boot-loader */

static libusb_context *uc = NULL;
static libusb_device_handle *rp_dev = NULL;
static int from_rp = 0, to_rp = 0;

#define LIBUSB_ASSERT(x) { int __val = (x); if (__val < 0) { \
	fprintf(stderr,"libusb: %s\n", libusb_strerror(__val)); return -1; } }
#define RET_ASSERT(x) { int __val = (x); if (__val < 0) { \
	fprintf(stderr,"failure!\n"); return -1; } }


static int
rp2040_open(void)
{
	LIBUSB_ASSERT(libusb_init(&uc))
	LIBUSB_ASSERT(libusb_set_option(uc, LIBUSB_OPTION_LOG_LEVEL,
		LIBUSB_LOG_LEVEL_WARNING))

	libusb_device **devs = NULL;
	int num_devs = libusb_get_device_list(uc, &devs);
	LIBUSB_ASSERT(num_devs)

	int i;
	for (i = 0; i < num_devs; i++) {
		struct libusb_device_descriptor dd;
		LIBUSB_ASSERT(libusb_get_device_descriptor(devs[i], &dd))
		if ((dd.idVendor != RP2040_VENDOR) ||
		    (dd.idProduct != RP2040_PRODUCT)) {
			continue;
		}

		struct libusb_config_descriptor *cd;
		LIBUSB_ASSERT(libusb_get_active_config_descriptor(devs[i], &cd))

		int j;			/* look for PICOBOOT vendor interface */
		for (j = 0; j < cd->bNumInterfaces; j++) {
			const struct libusb_interface *iface =
							&cd->interface[j];
			if (iface->num_altsetting < 1) continue;
			const struct libusb_interface_descriptor *id =
							&iface->altsetting[0];
			if ((id->bInterfaceClass != LIBUSB_CLASS_VENDOR_SPEC) ||
			    (id->bInterfaceSubClass != 0) ||
			    (id->bInterfaceProtocol != 0))
				continue;
			int k;
			for (k = 0; k < id->bNumEndpoints; k++) {
				const struct libusb_endpoint_descriptor *ep =
							&id->endpoint[k];
				if ((ep->bmAttributes &
						LIBUSB_TRANSFER_TYPE_MASK) !=
				    LIBUSB_ENDPOINT_TRANSFER_TYPE_BULK) {
					continue;
				}
				if (ep->bEndpointAddress & 0x80) {
					from_rp = ep->bEndpointAddress;
				} else {
					to_rp = ep->bEndpointAddress;
				}
			}
			if (from_rp && to_rp) {
				LIBUSB_ASSERT(libusb_open(devs[i], &rp_dev))
				libusb_free_device_list(devs, 1);
				libusb_set_auto_detach_kernel_driver(rp_dev, 1);
				LIBUSB_ASSERT(libusb_claim_interface(rp_dev, j));
				/* in case device was left in an error state: */
				libusb_clear_halt(rp_dev, from_rp);
				libusb_clear_halt(rp_dev, to_rp);
				return 0;
			}
		}
	}

	fprintf(stderr, "Couldn't find RP2040.\n");
	libusb_free_device_list(devs, 1);

	return -1;
}

static void
rp2040_close(void)
{
	if (rp_dev) {
		libusb_close(rp_dev);
		rp_dev = NULL;
	}
	if (uc) {
		libusb_exit(uc);
		uc = NULL;
	}
}

/* XXX - the bulk transfer may stall to indicate failure, in which case we'll
 * wish we'd set a timeout, and we can query it with some different
 * libusb_control_transfer() sort of interface */
static int
rp2040_transfer(int input, void *buf, int n)
{
	int ret = -1;
	LIBUSB_ASSERT(libusb_bulk_transfer(rp_dev, (input ? from_rp : to_rp),
						buf, n, &ret, /*timeout=*/0))
	return ret;
}

/* add a command block wrapper around the command block */
/* input is 1 iff the payload is from the RP2040 device to the host */
static int
rp2040_command(unsigned char cmd, void *args, int arglen,
		void *data, int data_len)
{
	unsigned char out[32] = { 0 };
	if (arglen > 16) { fprintf(stderr, "oversized args\n"); return -1; }
	tobuf_4byte_le(&out[0], 0x431fd10b);
	tobuf_4byte_le(&out[4], 0);		/* token used in replies */
	out[8] = cmd;
	out[9] = arglen;
	tobuf_4byte_le(&out[12], data_len);
	memcpy(&out[16], args, arglen);

	RET_ASSERT(rp2040_transfer(/*input=*/0, out, sizeof out))

	int input = !!(cmd & 0x80);
	if (data_len) {
		RET_ASSERT(rp2040_transfer(input, data, data_len))
	}

			/* ack is in opposite direction */
	unsigned t;
	RET_ASSERT(rp2040_transfer(!input, &t, 1))
	return 0;
}

static int
rp2040_exclusive_access(void)
{
	unsigned char arg[1] = { 1 /* exclusive */ };
	return rp2040_command(0x01, &arg, sizeof arg, NULL, 0);
}

static int
rp2040_reboot(unsigned entry)
{
	unsigned char arg[12];
	tobuf_4byte_le(&arg[0], entry);
/* Put the initial stack in the XIP SRAM, though hopefully it won't be used and
 * an explicit one will be set!  I experimentally verified that the stack
 * doesn't seem to be written to by the bootloader itself so it should be
 * possible to totally not use this memory. */
	tobuf_4byte_le(&arg[4], XIP_SRAM_START + XIP_SRAM_LEN);
	tobuf_4byte_le(&arg[8], 1000);	/* delay 1000ms, for USB disconnect */
	return rp2040_command(0x02, &arg, sizeof arg, NULL, 0);
}

/* addr and len must each be FLASH_SECTOR_SIZE aligned */
static int
rp2040_flash_erase(unsigned addr, unsigned len)
{
	unsigned char arg[8];
	tobuf_4byte_le(&arg[0], addr);
	tobuf_4byte_le(&arg[4], len);
	return rp2040_command(0x03, &arg, sizeof arg, NULL, 0);
}

static int
rp2040_read(unsigned addr, void *data, int len)
{
	unsigned char arg[8];
	tobuf_4byte_le(&arg[0], addr);
	tobuf_4byte_le(&arg[4], len);
	return rp2040_command(0x84, &arg, sizeof arg, data, len);
}

/* write to flash must be 256-byte aligned */
static int
rp2040_write(unsigned addr, void *data, int len)
{
	unsigned char arg[8];
	tobuf_4byte_le(&arg[0], addr);
	tobuf_4byte_le(&arg[4], len);
	return rp2040_command(0x05, &arg, sizeof arg, data, len);
}

static int
rp2040_exit_xip(void)
{
	return rp2040_command(0x06, NULL, 0, NULL, 0);
}


static unsigned
rp2040_crc32(unsigned char *in)
{
	unsigned len = 252;
	unsigned val = 0;
	unsigned inval = 0, inofs = 0, inbitofs = 8;
	while (1) {
		int xor_mask = ((val&0x80000000) ? 0x04c11db7 : 0);
		/* shift left by one, adding a new bit on the right */
		if (inbitofs == 8) {
			if (inofs >= len+4) {
				return val & 0xFFFFFFFF;
			} else if (inofs >= len) { /* append suffix */
				inval = 0;
			} else {
				inval = in[inofs];
				if (inofs < 4) { /* "initial value" */
					inval ^= 0xFF;
				}
			}
			inofs++;
			inbitofs = 0;
		}
		val = (val << 1) | ((inval >> 7)&1);
		inval <<= 1;
		inbitofs++;
		val ^= xor_mask;
	}
}

/* it's on flash, so re-compute the CRC, and fill it in if it wasn't already
 * there */
static void
install_crc(void)
{
	unsigned char *first = blocks[0];
	if (!first) return;		/* no second-stage boot loader? */
	unsigned crc = rp2040_crc32(first);
	unsigned char crcbuf[4];
	tobuf_4byte_le(crcbuf, crc);
	if (!memcmp(&first[252], "\x00\x00\x00\x00", 4)) {
		/* it is waiting to be filled in */
		fprintf(stderr, "installing CRC %02X%02X%02X%02X\n",
				crcbuf[0], crcbuf[1], crcbuf[2], crcbuf[3]);
		memcpy(&first[252], crcbuf, 4);
	} else if (memcmp(&first[252], crcbuf, 4)) {
		/* it is already present, but wrong?? */
		fprintf(stderr, "expecting CRC in FLASH boot loader, got %02X%02X%02X%02X, should be %02X%02X%02X%02X\n",
			first[252], first[253], first[254], first[255],
			crcbuf[0], crcbuf[1], crcbuf[2], crcbuf[3]);
	} else {
		/* it is already correct */
	}
}






static int
do_write(char *fn)
{
	FILE *f = fopen(fn, "rb");
	if (!f) {
		perror(fn);
		return -1;
	}

	if (elf_read_header(f)) {
		fprintf(stderr, "couldn't read ELF header\n");
		return -1;
	}
	if (elf_entry && !(elf_entry & 1)) {
		fprintf(stderr, "entry point in ELF header is %08X,\nbut expected it to be odd to mark thumb code\n", elf_entry);
		return -1;
	}
	if (IS_SRAM_ADDR(elf_entry)) {
		fprintf(stderr, "SRAM image\n");
	} else {
		fprintf(stderr, "FLASH image\n");
		if (elf_entry && (elf_entry != FLASH_START+1)) {
			fprintf(stderr, "ignoring ELF entry point of %08X, FLASH always boots at %08X\n", elf_entry, FLASH_START+1);
		}
		elf_entry = 0;
	}
	if (elf_read_phs(f)) {
		fprintf(stderr, "couldn't read ELF program headers\n");
		return -1;
	}
	if (accum_blocks(f)) {
		fprintf(stderr, "couldn't collect blocks\n");
		return -1;
	}

	if (!elf_entry) {	/* if it's booting from flash, compute crc */
		install_crc();
	}

	fprintf(stderr, "opening RP2040...");
	if (rp2040_open()) {
		fprintf(stderr, "couldn't init RP2040 PICOBOOT\n");
		return -1;
	}
	fprintf(stderr, "done.\n");
	if (rp2040_exclusive_access()) {
		fprintf(stderr, "RP2040 rejected exclusive PICOBOOT mode\n");
		return -1;
	}
	if (!elf_entry) {		/* FLASH needs exit_xip?? */
		rp2040_exit_xip();
	}

	int i;
	int count = 0;
	int sector_cleared = 0;
	for (i = 0; i < blocks_count; i++) {
		unsigned addr = blocks_start + i*BLKSZ;
		if (!(addr & (FLASH_SECTOR_SIZE-1))) {
			sector_cleared = 0;	/* first block in sector */
		}
		if (!blocks[i]) continue;

		if (!elf_entry && !sector_cleared) { /* must pre-clear flash */
			fprintf(stderr, "erasing %08X\r", addr);
			rp2040_flash_erase(addr, FLASH_SECTOR_SIZE);
			sector_cleared = 1;
		}

		fprintf(stderr, "writing %08X\r", addr);
		count++;
		if (rp2040_write(addr, blocks[i], BLKSZ)) {
			fprintf(stderr, "RP2040 rejected WRITE at %08X\n",addr);
			return -1;
		}
	}
	fprintf(stderr, "\ndone %d blocks\n", count);

	fprintf(stderr,"RP2040 reboot...");
	if (rp2040_reboot(elf_entry)) {
		fprintf(stderr, "RP2040 rejected REBOOT\n");
		return -1;
	}
	rp2040_close();
	fprintf(stderr, "done.\n");
	return 0;
}

static void
hexdump(unsigned addr, unsigned char *buf, unsigned len)
{
	while (len) {
		int n = ((len > 16) ? 16 : len);
		int i;
		printf("%08X ", addr);
		for (i = 0; i < n; i++) {
			printf("%s%02X", ((i&3) ? "" : " "), buf[i]);
		}
		for (; i < 16; i++) {
			printf("%s  ", ((i&3) ? "" : " "));
		}
		printf("  ");
		for (int i = 0; i < n; i++) {
			printf("%c", (isprint(buf[i]) ? buf[i] : '.'));
		}
		printf("\n");
		addr += n;
		buf += n;
		len -= n;
	}
}

static int
do_read(unsigned addr, unsigned len, int is_hex)
{
	if (rp2040_open()) {
		fprintf(stderr, "couldn't init RP2040 PICOBOOT\n");
		return -1;
	}
	if ((addr <= (FLASH_START+FLASH_LEN)) && ((addr+len) >= FLASH_START)) {
		rp2040_exit_xip();
	}

	while (len) {
		unsigned char buf[256] = { 0 };
		unsigned to_read = len;
		if (to_read > 256) {
			/* so far as I know, PICOBOOT protocol allows large
			 * reads but I'm picking this arbitrary limit */
			to_read = 256;
		}
		if (rp2040_read(addr, buf, to_read)) {
			fprintf(stderr, "RP2040 READ failed\n");
			return -1;
		}
		if (is_hex) {
			hexdump(addr, buf, to_read);
		} else {
			fwrite(buf, to_read, 1, stdout);
		}
		addr += to_read;
		len -= to_read;
	}

	rp2040_close();
	return 0;
}

static int
do_reboot(unsigned addr)
{
	if (rp2040_open()) {
		fprintf(stderr, "couldn't init RP2040 PICOBOOT\n");
		return -1;
	}
	fprintf(stderr, "sending REBOOT...");
	if (rp2040_reboot(addr)) {
		fprintf(stderr, "RP2040 rejected REBOOT\n");
		return -1;
	}
	rp2040_close();
	fprintf(stderr, "done.\n");
	return 0;
}

static int
do_erase(void)
{
	if (rp2040_open()) {
		fprintf(stderr, "couldn't init RP2040 PICOBOOT\n");
		return -1;
	}
	rp2040_exit_xip();
	fprintf(stderr, "sending FLASH_ERASE %08X...", FLASH_START);
	if (rp2040_flash_erase(FLASH_START, FLASH_SECTOR_SIZE)) {
		fprintf(stderr, "RP2040 rejected FLASH_ERASE\n");
		return -1;
	}
	rp2040_close();
	fprintf(stderr, "done.\n");
	return 0;
}

int
main(int argc, char **argv)
{
	int ret = -1;
	if (((argc == 3) || (argc == 4)) &&
		   (!strcmp(argv[1], "dump") || !strcmp(argv[1], "hdump"))) {
		ret = do_read(strtoul(argv[2], NULL, 0),
			((argc == 4) ? strtoul(argv[3], NULL, 0) : 256),
			(argv[1][0] == 'h'));
	} else if (((argc == 2) || (argc == 3)) && !strcmp(argv[1], "reboot")) {
		ret = do_reboot((argc == 3) ? strtoul(argv[2], NULL, 0) : 0);
	} else if ((argc == 2) && !strcmp(argv[1], "erase")) {
		ret = do_erase();
	} else if (argc == 2) {
		ret = do_write(argv[1]);
	} else {
		fprintf(stderr,
"RP2040 USB PICOBOOT tool\n"
"usage: %s <fn.elf>            write the given object to the rp2040\n"
"   or: %s dump <addr> [len]   dump contents of memory to stdout (binary)\n"
"   or: %s hdump <addr> [len]  dump contents of memory to stdout (hex)\n"
"   or: %s reboot [addr]       reboot (to flash or to optional address)\n"
"   or: %s erase               erase first flash sector (disable boot)\n"
"when providing an address, prefix it with \"0x\" for hex\n",
				argv[0], argv[0], argv[0], argv[0], argv[0]);
		ret = -1;
	}
	rp2040_close();		/* in case an error condition skipped it */
	return ret;
}
